from pathlib import Path

from django.conf import settings
from django.core import management
from django.core.management.utils import get_random_secret_key

from mayan.apps.appearance.literals import (
    COMMAND_NAME_APPEARANCE_PREPARE_STATIC
)
from mayan.apps.autoadmin.literals import COMMAND_NAME_AUTOADMIN_CREATE
from mayan.apps.dependencies.literals import COMMAND_NAME_DEPENDENCIES_INSTALL
from mayan.apps.smart_settings.literals import COMMAND_NAME_SETTINGS_SAVE
from mayan.settings.literals import (
    DEFAULT_SECRET_KEY, DEFAULT_USER_SETTINGS_FOLDER, SECRET_KEY_FILENAME,
    SYSTEM_DIR
)

from ..exceptions import BaseCommonException
from ..signals import (
    signal_perform_upgrade, signal_pre_initial_setup, signal_pre_upgrade,
    signal_post_initial_setup, signal_post_upgrade
)


class CommonAppManagementCommand:
    def __init__(self):
        self.path_media_root = Path(settings.MEDIA_ROOT)

        self.path_system = self.path_media_root / SYSTEM_DIR

        self.path_secret_key = self.path_system / SECRET_KEY_FILENAME

        self.path_user_settings = self.path_media_root / DEFAULT_USER_SETTINGS_FOLDER

    def do_create_user_settings_folder(self):
        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            self.path_user_settings.mkdir(exist_ok=True)

            # Touch media/settings/__init__.py
            (self.path_user_settings / '__init__.py').touch()

    def do_initial_setup(self, force=False, no_dependencies=False):
        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            # Create the media folder.
            try:
                self.path_media_root.mkdir()
            except FileExistsError as exception:
                if not force:
                    raise BaseCommonException(
                        'Existing media files. Backup, remove this folder, '
                        'and try again. Or use the --force argument.'
                    ) from exception

            # Touch media/__init__.py
            (self.path_media_root / '__init__.py').touch()

            self.do_create_user_settings_folder()

            # Create the media/system folder.
            try:
                self.path_system.mkdir()
            except FileExistsError as exception:
                if not force:
                    raise BaseCommonException(
                        'System folder already exists.'
                    ) from exception

            with self.path_secret_key.open(mode='w') as file_object:
                secret_key = get_random_secret_key()
                file_object.write(secret_key)

            settings.SECRET_KEY = secret_key

        if settings.SECRET_KEY == DEFAULT_SECRET_KEY:
            raise BaseCommonException(
                'SECRET_KEY value not set. If local storage is disabled, '
                'pass the SECRET_KEY via an environment variable. '
                'A SECRET_KEY value can be generated using the '
                '`common_generate_random_secret_key` command.'
            )

        signal_pre_initial_setup.send(sender=self)

        if not no_dependencies:
            self.do_install_dependencies()

        management.call_command(command_name=COMMAND_NAME_AUTOADMIN_CREATE)

        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            management.call_command(command_name=COMMAND_NAME_SETTINGS_SAVE)

        signal_post_initial_setup.send(sender=self)

    def do_install_dependencies(self):
        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            management.call_command(
                command_name=COMMAND_NAME_DEPENDENCIES_INSTALL
            )
            management.call_command(
                command_name=COMMAND_NAME_APPEARANCE_PREPARE_STATIC,
                interactive=False
            )

    def do_perform_upgrade(self, no_dependencies=False):
        try:
            signal_pre_upgrade.send(sender=self)
        except Exception as exception:
            raise BaseCommonException(
                'Error during signal_pre_upgrade signal.'
            ) from exception
        else:
            try:
                self.do_create_user_settings_folder()
            except FileExistsError:
                """
                This exception is expected when upgrading from recent version
                that already have a user settings folder.
                """
            else:
                if not no_dependencies:
                    self.do_install_dependencies()

                try:
                    signal_perform_upgrade.send(sender=self)
                except Exception as exception:
                    raise BaseCommonException(
                        'Error during signal_perform_upgrade signal.'
                    ) from exception
                else:
                    try:
                        signal_post_upgrade.send(sender=self)
                    except Exception as exception:
                        raise BaseCommonException(
                            'Error during signal_post_upgrade signal.'
                        ) from exception



import nh3

from django.utils.html import urlize
from django.utils.translation import gettext_lazy as _

from mayan.apps.templating.template_backends import Template


class MessageBusinessLogicMixin:
    def get_label(self):
        return Template(
            template_string='{{ instance.date_time }} @{{ instance.sender_object }} "{{ instance.subject }}"'
        ).render(
            context={'instance': self}
        )
    get_label.short_description = _(message='Label')

    def get_clean_body(self):
        output_linkified = urlize(
            autoescape=True, nofollow=False, text=self.body
        )

        output_cleaned = nh3.clean(
            html=output_linkified, link_rel='nofollow noopener noreferrer'
        )

        return output_cleaned

    def get_rendered_body(self):
        clean_body = self.get_clean_body()

        template = Template(template_string=clean_body)
        return template.render(
            context={'message': self}
        )

    def mark_read(self, user):
        self._event_actor = user
        self.read = True
        self.save(
            update_fields=('read',)
        )

    def mark_unread(self, user):
        self._event_actor = user
        self.read = False
        self.save(
            update_fields=('read',)
        )



from mayan.apps.dependencies.classes import PythonDependency
from mayan.apps.dependencies.environments import environment_testing

from mayan.settings.literals import PYTHON_PSUTIL_VERSION

PythonDependency(
    module=__name__, name='boto3', version_string='==1.40.24'
)
PythonDependency(
    module=__name__, name='django-storages', version_string='==1.14.6'
)
PythonDependency(
    module=__name__, name='extract-msg', version_string='==0.55.0'
)
PythonDependency(
    module=__name__, name='google-cloud-storage', version_string='==3.3.1'
)
PythonDependency(
    environment=environment_testing, module=__name__, name='psutil',
    version_string='=={}'.format(
        PYTHON_PSUTIL_VERSION
    )
)
PythonDependency(
    module=__name__, name='pycryptodome', version_string='==3.23.0'
)



from mayan.apps.icons.icons import Icon

icon_signature_captures = Icon(
    driver_name='fontawesome', symbol='file-contract'
)
icon_signature_capture_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='file-contract',
    secondary_symbol='plus'
)
icon_signature_capture_single_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_signature_capture_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_signature_capture_list = Icon(
    driver_name='fontawesome', symbol='file-contract'
)



from django.conf import settings
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave
from mayan.apps.templating.template_backends import Template

from .events import event_saved_resultset_created
from .managers import SavedResultsetEntryManager, SavedResultsetManager
from .model_mixins import SavedResultsetBusinessLogicModelMixin


class SavedResultset(
    ExtraDataModelMixin, SavedResultsetBusinessLogicModelMixin, models.Model
):
    _ordering_fields = ('timestamp', 'user')

    time_to_live = models.PositiveIntegerField(
        help_text=_(
            message='Time to keep the resultset in seconds. This value is '
            'increased every time the resultset is accessed.'
        ),
        verbose_name=_(message='Time to live')
    )
    timestamp = models.DateTimeField(
        db_index=True, help_text=_(
            message='The server date and time when the resultset was created.'
        ), verbose_name=_(message='Timestamp')
    )
    epoch = models.PositiveBigIntegerField(
        verbose_name=_('Epoch')
    )
    app_label = models.CharField(
        max_length=64, verbose_name=_(message='App label')
    )
    model_name = models.CharField(
        max_length=64, verbose_name=_(message='Model name')
    )
    search_query_text = models.TextField(
        verbose_name=_(message='Search query')
    )
    search_explainer_text = models.TextField(
        verbose_name=_(message='Search explainer text')
    )
    result_count = models.PositiveIntegerField(
        help_text=_(message='Number of results stored in the resultset.'),
        verbose_name=_(message='Result count')
    )
    user = models.ForeignKey(
        help_text=_(message='User for which the resultset was created.'),
        on_delete=models.CASCADE, related_name='saved_resultsets',
        to=settings.AUTH_USER_MODEL, verbose_name=_(message='User')
    )

    class Meta:
        ordering = ('-timestamp', 'user')
        verbose_name = _(message='Saved resultset')
        verbose_name_plural = _(message='Saved resultsets')

    objects = SavedResultsetManager()

    def __str__(self):
        model = self.get_model_class()
        model_verbose_name = model._meta.verbose_name

        return Template(
            template_string='{{ verbose_name }} - {{ timestamp }}'
        ).render(
            context={
                'timestamp': self.timestamp, 'verbose_name': model_verbose_name
            }
        )

    def get_absolute_url(self):
        return reverse(
            kwargs={'saved_resultset_id': self.pk},
            viewname='search:saved_resultset_result_list'
        )

    @method_event(
        created={
            'event': event_saved_resultset_created, 'target': 'self'
        }, event_manager_class=EventManagerSave
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)


class SavedResultsetEntry(models.Model):
    saved_resultset = models.ForeignKey(
        on_delete=models.CASCADE, related_name='entries', to=SavedResultset,
        verbose_name=_(message='Saved resultset')
    )
    object_id = models.BigIntegerField()

    class Meta:
        ordering = ('id',)
        verbose_name = _(message='Saved resultset entry')
        verbose_name_plural = _(message='Saved resultset entries')

    objects = SavedResultsetEntryManager()



from django.utils.translation import gettext_lazy as _

from rest_framework.reverse import reverse

from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import FilteredPrimaryKeyRelatedField

from ..models.workflow_transition_models import WorkflowTransition

from .workflow_template_state_serializers import (
    WorkflowTemplateStateSerializer
)


class WorkflowTemplateTransitionSerializer(
    serializers.HyperlinkedModelSerializer
):
    destination_state = WorkflowTemplateStateSerializer(
        label=_(message='Destination state'), read_only=True
    )
    destination_state_id = FilteredPrimaryKeyRelatedField(
        help_text=_(
            message='Primary key of the destination state to be added.'
        ), label=_(message='Destination state ID'),
        source_queryset_method='get_workflow_template_state_queryset',
        write_only=True
    )
    field_list_url = serializers.SerializerMethodField(
        label=_(message='Field list URL')
    )
    origin_state = WorkflowTemplateStateSerializer(
        label=_(message='Origin state'), read_only=True
    )
    origin_state_id = FilteredPrimaryKeyRelatedField(
        help_text=_(
            message='Primary key of the origin state to be added.'
        ), label=_(message='Origin state ID'),
        source_queryset_method='get_workflow_template_state_queryset',
        write_only=True
    )
    trigger_list_url = serializers.SerializerMethodField(
        label=_(message='Trigger list URL')
    )
    url = serializers.SerializerMethodField(
        label=_(message='URL')
    )
    workflow_template_id = serializers.IntegerField(
        label=_(message='Workflow template ID'), read_only=True,
        source='workflow_id'
    )
    workflow_template_url = serializers.SerializerMethodField(
        label=_(message='Workflow template URL')
    )

    class Meta:
        fields = (
            'condition', 'destination_state', 'destination_state_id',
            'field_list_url', 'id', 'label', 'origin_state',
            'origin_state_id', 'trigger_list_url', 'url',
            'workflow_template_id', 'workflow_template_url'
        )
        model = WorkflowTransition
        read_only_fields = (
            'field_list_url', 'id', 'trigger_list_url', 'url',
            'workflow_template_id', 'workflow_template_url'
        )

    def create(self, validated_data):
        validated_data['destination_state'] = validated_data.pop(
            'destination_state_id'
        )
        validated_data['origin_state'] = validated_data.pop(
            'origin_state_id'
        )

        return super().create(
            validated_data=validated_data
        )

    def get_field_list_url(self, instance):
        return reverse(
            format=self.context['format'], kwargs={
                'workflow_template_id': instance.workflow_id,
                'workflow_template_transition_id': instance.pk
            }, request=self.context['request'],
            viewname='rest_api:workflow-template-transition-field-list'
        )

    def get_trigger_list_url(self, instance):
        return reverse(
            format=self.context['format'], kwargs={
                'workflow_template_id': instance.workflow_id,
                'workflow_template_transition_id': instance.pk
            }, request=self.context['request'],
            viewname='rest_api:workflow-template-transition-trigger-list'
        )

    def get_url(self, instance):
        return reverse(
            format=self.context['format'], kwargs={
                'workflow_template_id': instance.workflow.pk,
                'workflow_template_transition_id': instance.pk
            }, request=self.context['request'],
            viewname='rest_api:workflow-template-transition-detail'
        )

    def get_workflow_template_state_queryset(self):
        return self.context['workflow_template'].states.all()

    def get_workflow_template_url(self, instance):
        return reverse(
            format=self.context['format'], kwargs={
                'workflow_template_id': instance.workflow.pk
            }, request=self.context['request'],
            viewname='rest_api:workflow-template-detail'
        )

    def update(self, instance, validated_data):
        validated_data['destination_state'] = validated_data.pop(
            'destination_state_id'
        )
        validated_data['origin_state'] = validated_data.pop(
            'origin_state_id'
        )

        return super().update(
            instance=instance, validated_data=validated_data
        )



from django.db import models

from .literals import DEFAULT_CREATE_BULK_BATCH_SIZE


class ManagerMinixCreateBulk(models.Manager):
    create_bulk_batch_size = DEFAULT_CREATE_BULK_BATCH_SIZE

    def create_bulk(self):
        batch = []
        count = 0

        try:
            while True:
                kwargs = (yield)
                model_instance = self.model(**kwargs)
                batch.append(model_instance)
                count += 1

                if count >= self.create_bulk_batch_size:
                    count = 0

                    self.bulk_create(
                        batch_size=self.create_bulk_batch_size, objs=batch
                    )
                    batch = []
        except GeneratorExit:
            self.bulk_create(
                batch_size=self.create_bulk_batch_size, objs=batch
            )



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link
from mayan.apps.navigation.utils import get_content_type_kwargs_factory

from .icons import (
    icon_global_error_log_entry_list, icon_object_error_log_entry_delete,
    icon_object_error_log_entry_list, icon_object_error_log_entry_list_clear
)
from .permissions import (
    permission_error_log_entry_delete, permission_error_log_entry_view
)

link_global_error_log_partition_entry_list = Link(
    icon=icon_global_error_log_entry_list, text=_(message='Global error log'),
    view='logging:global_error_log_partition_entry_list'
)
link_object_error_log_entry_delete = Link(
    icon=icon_object_error_log_entry_delete,
    kwargs={
        'app_label': 'resolved_object.error_log_partition.content_type.app_label',
        'model_name': 'resolved_object.error_log_partition.content_type.model',
        'object_id': 'resolved_object.error_log_partition.object_id',
        'error_log_partition_entry_id': 'resolved_object.pk'
    }, permission=permission_error_log_entry_delete, tags='dangerous',
    text=_(message='Delete'), view='logging:object_error_log_entry_delete'
)
link_object_error_log_entry_list = Link(
    icon=icon_object_error_log_entry_list,
    kwargs=get_content_type_kwargs_factory(variable_name='resolved_object'),
    permission=permission_error_log_entry_view, text=_(message='Errors'),
    view='logging:object_error_log_entry_list'
)
link_object_error_log_entry_list_clear = Link(
    icon=icon_object_error_log_entry_list_clear,
    kwargs=get_content_type_kwargs_factory(variable_name='resolved_object'),
    permission=permission_error_log_entry_delete, text=_(message='Clear errors'),
    view='logging:object_error_log_entry_list_clear'
)



import logging

import yaml

from django.core.exceptions import ValidationError
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.serialization import yaml_load
from mayan.apps.credentials.class_mixins import (
    BackendMixinCredentialsOptional
)
from mayan.apps.templating.template_backends import Template

from ..classes import SourceStoredFile

from .literals import (
    DEFAULT_STORAGE_BACKEND, DEFAULT_STORAGE_BACKEND_ARGUMENTS
)

logger = logging.getLogger(name=__name__)


class SourceBackendMixinStoredFileLocationStorageBackend(
    BackendMixinCredentialsOptional
):
    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields.update(
            {
                'storage_backend': {
                    'class': 'django.forms.CharField',
                    'default': DEFAULT_STORAGE_BACKEND,
                    'help_text': _(
                        message='Python path to the Storage subclass used to '
                        'access the source files.'
                    ),
                    'kwargs': {
                        'max_length': 255,
                    },
                    'label': _(message='Storage backend'),
                    'required': True
                },
                'storage_backend_arguments': {
                    'class': 'mayan.apps.templating.fields.TemplateField',
                    'default': DEFAULT_STORAGE_BACKEND_ARGUMENTS,
                    'kwargs': {
                        'initial_help_text': _(
                            message='Arguments to pass to the storage '
                            'backend. Use YAML format. The credential object '
                            'is available as {{ credential }}.'
                        ),
                        'max_length': 255,
                    },
                    'label': _(message='Storage backend arguments'),
                    'required': False
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='Storage'), {
                    'fields': (
                        'storage_backend', 'storage_backend_arguments'
                    )
                },
            ),
        )

        return fieldsets

    def get_storage_backend_arguments(self):
        field_value_arguments = self.kwargs.get(
            'storage_backend_arguments', '{}'
        )
        template_arguments = Template(template_string=field_value_arguments)
        context = {}

        credential = self.get_credential()
        if credential:
            context['credential'] = credential

        arguments = '{}'.format(
            template_arguments.render(context=context)
        )

        # Typecast `SafeString` to `str`.
        arguments = arguments.strip()

        try:
            return yaml_load(stream=arguments)
        except yaml.YAMLError:
            raise ValidationError(
                _(
                    message='Unable to initialize storage. Check the storage '
                    'backend arguments.'
                )
            )

    def get_storage_backend_class(self):
        try:
            return import_string(
                dotted_path=self.kwargs.get('storage_backend')
            )
        except Exception as exception:
            message = _(
                message='Unable to initialize storage. Check the storage '
                'backend dotted path.'
            )
            raise ValueError(message) from exception

    def get_storage_backend_instance(self):
        storage_backend_arguments = self.get_storage_backend_arguments()
        storage_backend_class = self.get_storage_backend_class()

        try:
            return storage_backend_class(**storage_backend_arguments)
        except Exception as exception:
            message = _(
                message='Unable to initialize storage; %s'
            ) % exception

            logger.fatal(message)
            raise TypeError(message) from exception

    def get_stored_file_list(self):
        regex_exclude = self.get_regex_exclude()
        regex_include = self.get_regex_include()

        storage_backend_instance = self.get_storage_backend_instance()

        try:
            # Specify '' with no argument name for compatibility. Django
            # requires a `path` argument while boto3 requires a `name`
            # argument.
            folders, entries = storage_backend_instance.listdir('')

            for entry in entries:
                entry_name = str(entry)

                if regex_include.match(string=entry_name) and not regex_exclude.match(string=entry_name):
                    yield SourceStoredFile(filename=entry_name, source=self)
        except Exception as exception:
            message = 'Unable get list of files from source: {}; {}'.format(
                self, exception
            )

            logger.error(message)
            raise ValueError(message) from exception



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0068_auto_20201024_1852')
    ]

    operations = [
        migrations.DeleteModel(
            name='DocumentFilePageResult'
        ),
        migrations.DeleteModel(
            name='DocumentVersionPageResult'
        ),
        migrations.CreateModel(
            name='DocumentFilePageSearchResult',
            fields=[],
            options={
                'verbose_name': 'Document file page',
                'verbose_name_plural': 'Document file pages',
                'ordering': ('document_file__document', 'page_number'),
                'proxy': True,
                'indexes': [],
                'constraints': []
            },
            bases=('documents.documentfilepage',)
        ),
        migrations.CreateModel(
            name='DocumentFileSearchResult',
            fields=[],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': []
            },
            bases=('documents.documentfile',)
        ),
        migrations.CreateModel(
            name='DocumentSearchResult',
            fields=[],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': []
            },
            bases=('documents.document',)
        ),
        migrations.CreateModel(
            name='DocumentVersionPageSearchResult',
            fields=[],
            options={
                'verbose_name': 'Document version page',
                'verbose_name_plural': 'Document version pages',
                'proxy': True,
                'indexes': [],
                'constraints': []
            },
            bases=('documents.documentversionpage',)
        ),
        migrations.CreateModel(
            name='DocumentVersionSearchResult',
            fields=[],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': []
            },
            bases=('documents.documentversion',)
        ),
        migrations.AlterField(
            model_name='document',
            name='label',
            field=models.CharField(
                blank=True, db_index=True, default='',
                help_text='A short text identifying the document. By '
                'default, will be set to the filename of the first file '
                'uploaded to the document.', max_length=255,
                verbose_name='Label'
            )
        ),
        migrations.AlterField(
            model_name='document',
            name='language',
            field=models.CharField(
                blank=True, default='eng',
                help_text='The primary language in the document.',
                max_length=8, verbose_name='Language'
            )
        ),
        migrations.AlterField(
            model_name='documentfile',
            name='filename',
            field=models.CharField(
                blank=True, max_length=255, verbose_name='Filename'
            )
        ),
        migrations.AlterField(
            model_name='documentversionpage',
            name='content_type',
            field=models.ForeignKey(
                help_text='Content type for the source object of the page.',
                on_delete=django.db.models.deletion.CASCADE,
                to='contenttypes.ContentType'
            )
        ),
        migrations.AlterField(
            model_name='documentversionpage',
            name='object_id',
            field=models.PositiveIntegerField(
                help_text='ID for the source object of the page.'
            )
        ),
        migrations.AlterField(
            model_name='documentversionpage',
            name='page_number',
            field=models.PositiveIntegerField(
                db_index=True, default=1,
                help_text='Unique integer number for the page. Pages are '
                'ordered by this number.', verbose_name='Page number'
            )
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='File metadata'), name='file_metadata'
)

event_file_metadata_document_file_submitted = namespace.add_event_type(
    label=_(message='Document file submitted for file metadata processing'),
    name='document_file_submitted'
)
event_file_metadata_document_file_finished = namespace.add_event_type(
    label=_(message='Document file file metadata processing finished'),
    name='document_file_finished'
)



import logging

from django.db import models

from mayan.apps.databases.manager_mixins import ManagerMinixCreateBulk

logger = logging.getLogger(name=__name__)


class FileMetadataEntryManager(ManagerMinixCreateBulk, models.Manager):
    """
    Leave empty as no other methods are necessary at the moment. This is just
    to allow the mixin into the manager class.
    """


class ModelManagerDocumentTypeDriverConfigurationValid(models.Manager):
    def get_queryset(self):
        queryset = super().get_queryset()

        return queryset.filter(stored_driver__exists=True)


class ModelManagerStoredDriverValid(models.Manager):
    def get_queryset(self):
        queryset = super().get_queryset()

        return queryset.filter(exists=True)



from mayan.apps.storage.download_backends.http import (
    DownloadBackendFileDirectStorage
)

from .mixins import DownloadBackendMixinDocumentFile


class DownloadBackendDocumentFileFileDirectStorage(
    DownloadBackendMixinDocumentFile, DownloadBackendFileDirectStorage
):
    def get_download_file_object(self, obj):
        return obj.open(raw=True)

    def get_download_mime_type_and_encoding(self, obj):
        return (obj.mimetype, obj.encoding)



from django.urls import re_path

from .api_views.document_file_api_views import (
    APIDocumentFileMetadataDriverDetailView,
    APIDocumentFileMetadataDriverListView,
    APIDocumentFileMetadataEntryDetailView,
    APIDocumentFileMetadataEntryListView, APIDocumentFileMetadataSubmitView
)
from .api_views.document_type_api_views import (
    APIDocumentTypeFileMetadataDriverConfigurationDetailView,
    APIDocumentTypeFileMetadataDriverConfigurationListView
)
from .api_views.file_metadata_api_views import (
    APIStoredDriverDetailView, APIStoredDriverListView
)
from .views.document_file_views import (
    DocumentFileMetadataDriverAttributeListView,
    DocumentFileMetadataDriverListView, DocumentFileMetadataSubmitView
)
from .views.document_type_views import (
    DocumentTypeFileMetadataDriverConfigurationEditView,
    DocumentTypeFileMetadataDriverConfigurationListView,
    DocumentTypeFileMetadataSubmitView,
)
from .views.tool_views import FileMetadataDriverListView

urlpatterns_document_types = [
    re_path(
        route=r'^document_types/(?P<document_type_id>\d+)/file_metadata/drivers/$',
        name='document_type_file_metadata_driver_configuration_list',
        view=DocumentTypeFileMetadataDriverConfigurationListView.as_view()
    ),
    re_path(
        route=r'^document_types/(?P<document_type_id>\d+)/file_metadata/drivers/(?P<stored_driver_id>\d+)/edit/$',
        name='document_type_file_metadata_driver_configuration_edit',
        view=DocumentTypeFileMetadataDriverConfigurationEditView.as_view()
    ),
    re_path(
        route=r'^document_types/submit/$',
        name='document_type_file_metadata_submit',
        view=DocumentTypeFileMetadataSubmitView.as_view()
    )
]

urlpatterns_documents = [
    re_path(
        route=r'^documents/files/drivers/(?P<document_file_driver_id>\d+)/attributes/$',
        name='document_file_metadata_driver_attribute_list',
        view=DocumentFileMetadataDriverAttributeListView.as_view()
    ),
    re_path(
        route=r'^documents/files/(?P<document_file_id>\d+)/drivers/$',
        name='document_file_metadata_driver_list',
        view=DocumentFileMetadataDriverListView.as_view()
    ),
    re_path(
        route=r'^documents/files/(?P<document_file_id>\d+)/submit/$',
        name='document_file_metadata_single_submit',
        view=DocumentFileMetadataSubmitView.as_view()
    ),
    re_path(
        route=r'^documents/files/multiple/submit/$',
        name='document_file_metadata_multiple_submit',
        view=DocumentFileMetadataSubmitView.as_view()
    )
]

urlpatterns_drivers = [
    re_path(
        route=r'^driver/list/$', name='file_metadata_driver_list',
        view=FileMetadataDriverListView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_document_types)
urlpatterns.extend(urlpatterns_documents)
urlpatterns.extend(urlpatterns_drivers)

api_urls_document_files = [
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/files/(?P<document_file_id>[0-9]+)/file_metadata/drivers/$',
        name='document_file_metadata_driver-list',
        view=APIDocumentFileMetadataDriverListView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/files/(?P<document_file_id>[0-9]+)/file_metadata/drivers/(?P<driver_id>[0-9]+)/$',
        name='document_file_metadata_driver-detail',
        view=APIDocumentFileMetadataDriverDetailView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/files/(?P<document_file_id>[0-9]+)/file_metadata/drivers/(?P<driver_id>[0-9]+)/entries/$',
        name='document_file_metadata_entry-list',
        view=APIDocumentFileMetadataEntryListView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/files/(?P<document_file_id>[0-9]+)/file_metadata/drivers/(?P<driver_id>[0-9]+)/entries/(?P<entry_id>[0-9]+)/$',
        name='document_file_metadata_entry-detail',
        view=APIDocumentFileMetadataEntryDetailView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/files/(?P<document_file_id>[0-9]+)/file_metadata/submit/$',
        name='document_file_metadata-submit',
        view=APIDocumentFileMetadataSubmitView.as_view()
    )
]

api_urls_document_types = [
    re_path(
        route=r'^document_types/(?P<document_type_id>[0-9]+)/file_metadata/drivers/$',
        name='document_type_file_metadata_configuration-list',
        view=APIDocumentTypeFileMetadataDriverConfigurationListView.as_view()
    ),
    re_path(
        route=r'^document_types/(?P<document_type_id>[0-9]+)/file_metadata/drivers/(?P<driver_id>[0-9]+)/$',
        name='document_type_file_metadata_configuration-detail',
        view=APIDocumentTypeFileMetadataDriverConfigurationDetailView.as_view()
    )
]

api_urls_file_metadata = [
    re_path(
        route=r'^file_metadata_drivers/$', name='file_metadata_driver-list',
        view=APIStoredDriverListView.as_view()
    ),
    re_path(
        route=r'^file_metadata_drivers/(?P<stored_driver_id>[0-9]+)/$',
        name='file_metadata_driver-detail',
        view=APIStoredDriverDetailView.as_view()
    )
]

api_urls = []
api_urls.extend(api_urls_document_files)
api_urls.extend(api_urls_document_types)
api_urls.extend(api_urls_file_metadata)



import logging

from django.apps import apps

from mayan.celery import app

logger = logging.getLogger(name=__name__)


@app.task(ignore_result=True)
def task_document_type_document_stubs_delete():
    DocumentType = apps.get_model(
        app_label='documents', model_name='DocumentType'
    )

    logger.debug(msg='Starting')

    DocumentType.objects.document_stubs_delete()

    logger.debug(msg='Finished')


@app.task(ignore_result=True)
def task_document_type_document_trash_periods_check():
    DocumentType = apps.get_model(
        app_label='documents', model_name='DocumentType'
    )

    DocumentType.objects.check_trash_periods()


@app.task(ignore_result=True)
def task_document_type_trashed_document_delete_periods_check():
    DocumentType = apps.get_model(
        app_label='documents', model_name='DocumentType'
    )

    DocumentType.objects.check_delete_periods()



from django.http import Http404
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.views.generics import (
    MultipleObjectDeleteView, SingleObjectListView
)
from mayan.apps.views.view_mixins import (
    ExternalObjectViewMixin, ViewMixinExternalObjectOwnerPlusFilteredQueryset,
    ViewMixinOwnerPlusFilteredQueryset
)

from ..icons import (
    icon_saved_resultset_delete_single, icon_saved_resultset_list,
    icon_saved_resultset_result_list,
)
from ..models import SavedResultset
from ..permissions import (
    permission_saved_resultset_delete, permission_saved_resultset_view
)
from ..search_interpreters import SearchInterpreter

from .search_views import SearchResultView


class SavedResultsetDeleteView(
    ViewMixinOwnerPlusFilteredQueryset, MultipleObjectDeleteView
):
    error_message = _(
        message='Error deleting saved resultset "%(instance)s"; %(exception)s'
    )
    model = SavedResultset
    object_optional_permission = permission_saved_resultset_delete
    pk_url_kwarg = 'saved_resultset_id'
    post_action_redirect = reverse_lazy(
        viewname='search:saved_resultset_list'
    )
    success_message_plural = _(
        message='%(count)d saved resultsets deleted successfully.'
    )
    success_message_single = _(
        message='Saved resultset "%(object)s" deleted successfully.'
    )
    success_message_singular = _(
        message='%(count)d saved resultset deleted successfully.'
    )
    title_plural = _(
        message='Delete the %(count)d selected saved resultsets.'
    )
    title_single = _(message='Delete saved resultset: %(object)s.')
    title_singular = _(message='Delete the %(count)d saved resultset.')
    view_icon = icon_saved_resultset_delete_single


class SavedResultsetListView(
    ViewMixinOwnerPlusFilteredQueryset, SingleObjectListView
):
    model = SavedResultset
    object_optional_permission = permission_saved_resultset_view
    view_icon = icon_saved_resultset_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_saved_resultset_list,
            'no_results_text': _(
                message='Saved resultsets are objects that store the results '
                'of a previous search for a determined amount of time.'
            ),
            'no_results_title': _(message='No saved resultsets available'),
            'title': _(message='Saved resultsets')
        }


class SavedResultsetResultListView(
    ViewMixinExternalObjectOwnerPlusFilteredQueryset, ExternalObjectViewMixin,
    SearchResultView
):
    external_object_class = SavedResultset
    external_object_optional_permission = permission_saved_resultset_view
    external_object_pk_url_kwarg = 'saved_resultset_id'
    view_icon = icon_saved_resultset_result_list

    def get_extra_context(self):
        context = super().get_extra_context()

        context.update(
            {'object': self.external_object}
        )

        return context

    def get_search_model(self):
        try:
            return self.external_object.get_search_model()
        except KeyError as exception:
            raise Http404(
                str(exception)
            )

    def get_search_queryset(self):
        query = self.external_object.get_search_query_dict()

        self.search_interpreter = SearchInterpreter.init(
            query=query, search_model=self.search_model
        )

        return self.external_object.get_result_queryset()



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0091_fix_documenttype_verbose_name'),
        ('document_states', '0039_workflow_ignore_completed')
    ]

    operations = [
        migrations.AlterField(
            field=models.ForeignKey(
                help_text='The document to which the workflow instance is '
                'attached.', on_delete=django.db.models.deletion.CASCADE,
                related_name='workflows', to='documents.document',
                verbose_name='Document'
            ), model_name='workflowinstance', name='document'
        ),
        migrations.AlterField(
            field=models.ForeignKey(
                help_text='The currently active state of the workflow '
                'instance.', on_delete=django.db.models.deletion.CASCADE,
                related_name='workflow_instances',
                to='document_states.workflowstate',
                verbose_name='Active state'
            ),
            model_name='workflowinstance', name='state_active'
        ),
        migrations.AlterField(
            field=models.ForeignKey(
                help_text='The workflow template.',
                on_delete=django.db.models.deletion.CASCADE,
                related_name='instances', to='document_states.workflow',
                verbose_name='Workflow'
            ), model_name='workflowinstance', name='workflow'
        )
    ]



from mayan.apps.rest_api import generics

from .permissions import (
    permission_message_create, permission_message_delete,
    permission_message_edit, permission_message_view
)
from .serializers import MessageSerializer


class APIMessageDetailView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected message.
    get: Return the details of the selected message.
    patch: Partially edit the selected message.
    put: Edit the selected message.
    """
    lookup_url_kwarg = 'message_id'
    mayan_object_permission_map = {
        'GET': permission_message_view,
        'PUT': permission_message_edit,
        'PATCH': permission_message_edit,
        'DELETE': permission_message_delete
    }
    serializer_class = MessageSerializer

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_source_queryset(self):
        return self.request.user.messages.all()


class APIMessageListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the messages.
    post: Create a new message.
    """
    mayan_object_permission_map = {'GET': permission_message_view}
    mayan_view_permission_map = {'POST': permission_message_create}
    serializer_class = MessageSerializer

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_source_queryset(self):
        return self.request.user.messages.all()



from django.template import Library

from ..literals import (
    FILTER_PREFIX, QUERY_PARAMETER_ANY_FIELD, SEARCH_MODEL_NAME_KWARG
)
from ..search_models import SearchModel

register = Library()


@register.simple_tag(name='search_get_default_search_model')
def tag_search_get_default_search_model():
    return SearchModel.get_default()


@register.simple_tag(name='search_get_filter_prefix')
def tag_search_get_filter_prefix():
    return FILTER_PREFIX


@register.simple_tag(name='search_get_query_any_field_value')
def tag_search_get_query_any_field_value(query=None):
    query = query or {}
    return query.get(QUERY_PARAMETER_ANY_FIELD, '')


@register.simple_tag(name='search_get_query_parameter_any_field')
def tag_search_get_query_parameter_any_field():
    return QUERY_PARAMETER_ANY_FIELD


@register.simple_tag(name='search_get_search_models')
def tag_search_get_search_models():
    return SearchModel.all()


@register.simple_tag(name='search_get_search_model_name_query_variable')
def tag_search_get_search_model_name_query_variable():
    return '_{}'.format(SEARCH_MODEL_NAME_KWARG)



import errno
import logging
import os
from shutil import copyfile
import sys

import yaml

from django.apps import apps
from django.conf import settings
from django.db.utils import OperationalError, ProgrammingError

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.apps.common.serialization import yaml_dump, yaml_load

from .classes import Setting
from .exceptions import SettingsException
from .literals import (
    COMMAND_NAME_SETTINGS_REVERT, SMART_SETTINGS_NAMESPACES_NAME
)
from .setting_namespaces import SettingNamespace

logger = logging.getLogger(name=__name__)


class SettingCluster(AppsModuleLoaderMixin):
    _loader_module_name = 'settings'

    @staticmethod
    def read_configuration_file(filepath):
        try:
            with open(file=filepath) as file_object:
                file_object.seek(0, os.SEEK_END)
                if file_object.tell():
                    file_object.seek(0)
                    try:
                        return yaml_load(stream=file_object)
                    except yaml.YAMLError as exception:
                        exit(
                            'Error loading configuration file: {}; {}'.format(
                                filepath, exception
                            )
                        )
        except IOError as exception:
            if exception.errno == errno.ENOENT:
                # No config file, return empty dictionary.
                return {}
            else:
                raise

    def __init__(self, name):
        self.configuration_file_cache = None
        self.name = name
        self.namespace_dict = {}
        self.setting_dict = {}

    def do_cache_invalidate(self):
        for namespace in self.namespace_dict.values():
            namespace.do_cache_invalidate()

        self.configuration_file_cache = None

    def do_configuration_file_revert(self):
        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            try:
                copyfile(
                    src=settings.CONFIGURATION_LAST_GOOD_FILEPATH,
                    dst=settings.CONFIGURATION_FILEPATH
                )
            except IOError as exception:
                if exception.errno == errno.ENOENT:
                    raise SettingsException(
                        'There is no last valid version to restore.'
                    ) from exception
                else:
                    raise
        else:
            logger.info(
                'Local storage is disabled, cannot revert not existing '
                'configuration.'
            )

    def do_configuration_file_save(self, path=None):
        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            if not path:
                path = settings.CONFIGURATION_FILEPATH

            try:
                with open(file=path, mode='w') as file_object:
                    file_object.write(
                        self.get_data_dump()
                    )
            except IOError as exception:
                if exception.errno == errno.ENOENT:
                    logger.warning(
                        'The path to the configuration file `%s` doesn\'t '
                        'exist. It is not possible to save the backup file.',
                        path
                    )
        else:
            logger.info(
                'Local storage is disabled, skip saving configuration.'
            )

    def do_last_known_good_save(self):
        # Don't write over the last good configuration if we are trying
        # to restore the last good configuration.
        if COMMAND_NAME_SETTINGS_REVERT not in sys.argv and not settings.CONFIGURATION_FILE_IGNORE:
            self.do_configuration_file_save(
                path=settings.CONFIGURATION_LAST_GOOD_FILEPATH
            )

    def do_namespace_add(self, **kwargs):
        setting_namespace = SettingNamespace(cluster=self, **kwargs)

        return setting_namespace

    def do_namespace_remove(self, name):
        setting_namespace = self.get_namespace(name=name)

        self.namespace_dict.pop(setting_namespace.name)

    def do_post_edit_function_call(self):
        ContentType = apps.get_model(
            app_label='contenttypes', model_name='ContentType'
        )

        for namespace in self.namespace_dict.values():
            namespace.do_post_edit_function_call()

        # Clear the content type cache to avoid the event system from trying
        # to use the same ID that were cached when the setting post edit
        # functions executed. This is because the settings execute before
        # the apps objects are created.
        ContentType.objects.clear_cache()

    def do_settings_updated_clear(self):
        UpdatedSetting = apps.get_model(
            app_label='smart_settings', model_name='UpdatedSetting'
        )

        queryset = UpdatedSetting.objects.all()

        try:
            queryset.delete()
        except (OperationalError, ProgrammingError):
            """
            Non fatal. Non initialized installation. Ignore exception.
            """

    def get_configuration_file_content(self):
        if settings.CONFIGURATION_FILE_IGNORE:
            return {}
        else:
            # Cache content the of the configuration file to speed up
            # initial boot up.
            if not self.configuration_file_cache:
                self.configuration_file_cache = SettingCluster.read_configuration_file(
                    filepath=settings.CONFIGURATION_FILEPATH
                ) or {}
            return self.configuration_file_cache

    def get_data_dump(self, filter_term=None, namespace_name=None):
        UpdatedSetting = apps.get_model(
            app_label='smart_settings', model_name='UpdatedSetting'
        )

        dictionary = {}

        if not namespace_name:
            namespace_dictionary = {}
            for namespace in self.get_namespace_list():
                namespace_dictionary[namespace.name] = {
                    'version': namespace.version
                }

            dictionary[SMART_SETTINGS_NAMESPACES_NAME] = namespace_dictionary

        if namespace_name:
            namespace_list = (
                self.get_namespace(name=namespace_name),
            )
        else:
            namespace_list = self.get_namespace_list()

        for namespace in namespace_list:
            for setting in namespace.get_setting_list():
                # If a namespace is specified, filter the list by that
                # namespace otherwise return always True to include all
                # (or not None == True).
                if (filter_term and filter_term.lower() in setting.global_name.lower()) or not filter_term:
                    try:
                        updated_setting = UpdatedSetting.objects.get(
                            global_name=setting.global_name
                        )
                    except (OperationalError, ProgrammingError, UpdatedSetting.DoesNotExist):
                        expressed_value = Setting.express_promises(
                            value=setting.value
                        )
                        dictionary[setting.global_name] = expressed_value
                    else:
                        dictionary[setting.global_name] = updated_setting.value_new

        return yaml_dump(data=dictionary, default_flow_style=False)

    def get_is_changed(self):
        UpdatedSetting = apps.get_model(
            app_label='smart_settings', model_name='UpdatedSetting'
        )

        try:
            return UpdatedSetting.objects.exists()
        except (OperationalError, ProgrammingError):
            return False

    def get_namespace(self, name):
        return self.namespace_dict[name]

    def get_namespace_configuration(self, name):
        namespace_configuration_map = self.get_namespace_configuration_map()

        return namespace_configuration_map.get(
            name, {}
        )

    def get_namespace_configuration_map(self):
        configuration_file_content = self.get_configuration_file_content()

        return configuration_file_content.get(
            SMART_SETTINGS_NAMESPACES_NAME, {}
        )

    def get_namespace_list(self):
        return sorted(
            self.namespace_dict.values(), key=lambda x: x.label
        )

    def get_setting(self, global_name):
        return self.setting_dict[global_name]

    def get_setting_list(self):
        return self.setting_dict.values()



import hashlib

from django.template.response import TemplateResponse
from django.utils.translation import gettext_lazy as _
from django.urls import reverse

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.common.menus import menu_list_facet

from .links import link_object_template_sandbox
from .permissions import permission_template_sandbox
from .template_backends import Template


class AJAXTemplate:
    _registry = {}

    @classmethod
    def all(cls, rendered=False, request=None):
        if not rendered:
            return cls._registry.values()
        else:
            result = []
            for template in cls._registry.values():
                result.append(
                    template.render(request=request)
                )
            return result

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    def __init__(self, name, template_name, context=None):
        self.context = context or None
        self.name = name
        self.template_name = template_name
        self.__class__._registry[name] = self

    def get_absolute_url(self):
        return reverse(
            kwargs={'name': self.name}, viewname='rest_api:template-detail'
        )

    def render(self, request):
        result = TemplateResponse(
            context=self.context, request=request,
            template=self.template_name
        ).render()

        # Calculate the hash of the bytes version but return the unicode
        # version.
        self.html = result.rendered_content.replace('\n', '')
        self.hex_hash = hashlib.sha256(string=result.content).hexdigest()
        return self


class ModelTemplating:
    _registry = {}

    @classmethod
    def do_render(cls, obj, template_string):
        model = obj._meta.model

        model_templating = cls.get_for_model(model=model)

        template = Template(template_string=template_string)
        result = template.render(
            context={
                model_templating.variable_name: obj
            }
        )

        return result

    @classmethod
    def get_for_model(cls, model):
        try:
            return cls._registry[model]
        except KeyError:
            raise KeyError(
                _(
                    message='Model `%s` is not available for template '
                    'sandbox.'
                ) % model._meta.verbose_name
            )

    def __init__(self, model, variable_name):
        self.model = model
        self.variable_name = variable_name

        ModelPermission.register(
            model=self.model, permissions=(permission_template_sandbox,)
        )

        menu_list_facet.bind_links(
            links=(link_object_template_sandbox,), sources=(self.model,)
        )

        self.__class__._registry[self.model] = self



from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '__first__'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL)
    ]

    operations = [
        migrations.CreateModel(
            name='Workflow',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'label', models.CharField(
                        unique=True, max_length=255, verbose_name='Label'
                    )
                ),
                (
                    'document_types', models.ManyToManyField(
                        related_name='workflows',
                        verbose_name='Document types',
                        to='documents.DocumentType'
                    )
                )
            ],
            options={
                'verbose_name': 'Workflow',
                'verbose_name_plural': 'Workflows'
            },
            bases=(models.Model,)
        ),
        migrations.CreateModel(
            name='WorkflowInstance',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False, auto_created=True,
                        primary_key=True
                    )
                ),
                (
                    'document', models.ForeignKey(
                        on_delete=models.CASCADE, related_name='workflows',
                        to='documents.Document', verbose_name='Document'
                    )
                ),
                (
                    'workflow', models.ForeignKey(
                        on_delete=models.CASCADE, related_name='instances',
                        to='document_states.Workflow',
                        verbose_name='Workflow'
                    )
                )
            ],
            options={
                'verbose_name': 'Workflow instance',
                'verbose_name_plural': 'Workflow instances'
            },
            bases=(models.Model,)
        ),
        migrations.CreateModel(
            name='WorkflowInstanceLogEntry',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False, auto_created=True,
                        primary_key=True
                    )
                ),
                (
                    'datetime', models.DateTimeField(
                        auto_now_add=True, verbose_name='Datetime',
                        db_index=True
                    )
                ),
                (
                    'comment', models.TextField(
                        verbose_name='Comment', blank=True
                    )
                )
            ],
            options={
                'verbose_name': 'Workflow instance log entry',
                'verbose_name_plural': 'Workflow instance log entries'
            },
            bases=(models.Model,)
        ),
        migrations.CreateModel(
            name='WorkflowState',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False, auto_created=True,
                        primary_key=True
                    )
                ),
                (
                    'label', models.CharField(
                        max_length=255, verbose_name='Label'
                    )
                ),
                (
                    'initial', models.BooleanField(
                        default=False, help_text='Select if this will be the '
                        'state with which you want the workflow to start in. '
                        'Only one state can be the initial state.',
                        verbose_name='Initial'
                    )
                ),
                (
                    'workflow', models.ForeignKey(
                        on_delete=models.CASCADE, related_name='states',
                        to='document_states.Workflow',
                        verbose_name='Workflow'
                    )
                )
            ],
            options={
                'verbose_name': 'Workflow state',
                'verbose_name_plural': 'Workflow states'
            },
            bases=(models.Model,)
        ),
        migrations.CreateModel(
            name='WorkflowTransition',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False, auto_created=True,
                        primary_key=True
                    )
                ),
                (
                    'label', models.CharField(
                        max_length=255, verbose_name='Label'
                    )
                ),
                (
                    'destination_state', models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name='destination_transitions',
                        to='document_states.WorkflowState',
                        verbose_name='Destination state'
                    )
                ),
                (
                    'origin_state', models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name='origin_transitions',
                        to='document_states.WorkflowState',
                        verbose_name='Origin state'
                    )
                ),
                (
                    'workflow', models.ForeignKey(
                        on_delete=models.CASCADE, related_name='transitions',
                        to='document_states.Workflow',
                        verbose_name='Workflow'
                    )
                )
            ],
            options={
                'verbose_name': 'Workflow transition',
                'verbose_name_plural': 'Workflow transitions'
            },
            bases=(models.Model,)
        ),
        migrations.AlterUniqueTogether(
            name='workflowtransition',
            unique_together={
                ('workflow', 'label', 'origin_state', 'destination_state')
            }
        ),
        migrations.AlterUniqueTogether(
            name='workflowstate',
            unique_together={
                ('workflow', 'label')
            }
        ),
        migrations.AddField(
            model_name='workflowinstancelogentry',
            name='transition',
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                to='document_states.WorkflowTransition',
                verbose_name='Transition'
            ),
            preserve_default=True
        ),
        migrations.AddField(
            model_name='workflowinstancelogentry',
            name='user',
            field=models.ForeignKey(
                on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL,
                verbose_name='User'
            ),
            preserve_default=True
        ),
        migrations.AddField(
            model_name='workflowinstancelogentry',
            name='workflow_instance',
            field=models.ForeignKey(
                on_delete=models.CASCADE, related_name='log_entries',
                to='document_states.WorkflowInstance',
                verbose_name='Workflow instance'
            ),
            preserve_default=True
        ),
        migrations.AlterUniqueTogether(
            name='workflowinstance',
            unique_together={
                ('document', 'workflow')
            }
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Sources setup'), name='sources_setup'
)

# Documents

permission_document_file_sources_metadata_view = namespace.add_permission(
    label=_(message='View document file source metadata'),
    name='source_metadata_view'
)

# Sources

permission_sources_create = namespace.add_permission(
    label=_(message='Create new document sources'),
    name='sources_setup_create'
)
permission_sources_delete = namespace.add_permission(
    label=_(message='Delete document sources'), name='sources_setup_delete'
)
permission_sources_edit = namespace.add_permission(
    label=_(message='Edit document sources'), name='sources_setup_edit'
)
permission_sources_view = namespace.add_permission(
    label=_(message='View existing document sources'),
    name='sources_setup_view'
)



import email
import logging
import random

from django.core.files.base import ContentFile
from django.utils.encoding import force_bytes
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.utils import convert_to_internal_name
from mayan.apps.credentials.class_mixins import BackendMixinCredentials
from mayan.apps.source_periodic.source_backends.mixins import (
    SourceBackendMixinPeriodicCompressed
)

from ..source_backend_actions import SourceBackendActionEmailDocumentUpload

logger = logging.getLogger(name=__name__)


class SourceBackendMixinEmail(
    BackendMixinCredentials, SourceBackendMixinPeriodicCompressed
):
    action_class_list = (SourceBackendActionEmailDocumentUpload,)

    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields.update(
            {
                'host': {
                    'class': 'django.forms.CharField',
                    'label': _(message='Host'),
                    'kwargs': {
                        'max_length': 128
                    },
                    'required': True
                },
                'ssl': {
                    'class': 'django.forms.BooleanField',
                    'default': True,
                    'label': _(message='SSL'),
                    'required': False
                },
                'port': {
                    'class': 'django.forms.IntegerField',
                    'help_text': _(
                        message='Typical choices are 110 for POP3, 995 for '
                        'POP3 over SSL, 143 for IMAP, 993 for IMAP over SSL.'
                    ),
                    'kwargs': {
                        'min_value': 0
                    },
                    'label': _(message='Port')
                },
                'store_body': {
                    'class': 'django.forms.BooleanField',
                    'default': True,
                    'help_text': _(
                        message='Store the body of the email as a text '
                        'document.'
                    ),
                    'label': _(message='Store email body'),
                    'required': False
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='Common email options'), {
                    'fields': (
                        'host', 'ssl', 'port', 'store_body'
                    )
                },
            ),
        )

        return fieldsets

    def process_message(self, message):
        bytes_message = force_bytes(s=message)

        message = email.message_from_bytes(
            policy=email.policy.default, s=bytes_message
        )

        return self._process_message_content(message=message)

    def _process_message_content(self, message, source_metadata=None):
        counter = 1
        if source_metadata is None:
            _source_metadata = {}
        else:
            _source_metadata = source_metadata.copy()

        for key, value in message.items():
            internal_name = convert_to_internal_name(value=key)
            _source_metadata[
                'email_{}'.format(internal_name)
            ] = value

        # Messages are tree based, do nested processing of message parts until
        # a message with no children is found, then work our way up.
        if message.is_multipart():
            for part in message.iter_parts():
                yield from self._process_message_content(
                    message=part, source_metadata=_source_metadata
                )
        else:
            # Treat inlines as attachments, both are extracted and saved as
            # documents.
            if message.is_attachment() or message.get_content_disposition() == 'inline':
                content = message.get_content()
                if len(content) != 0:
                    detected_filename = message.get_filename()
                    if detected_filename:
                        label = detected_filename
                    else:
                        label = 'attachment-{}'.format(counter)
                        counter += 1

                    yield {
                        'file': ContentFile(
                            content=content, name=label,
                        ), 'source_metadata': _source_metadata
                    }
            else:
                # If it is not an attachment then it should be a body message
                # part.
                if message.get_content_type() == 'text/html':
                    label = 'email_body.html'
                else:
                    label = 'email_body.txt'

                if self.kwargs['store_body']:
                    content = message.get_content()
                    bytes_content = force_bytes(s=content)
                    yield {
                        'file': ContentFile(
                            content=bytes_content, name=label
                        ), 'source_metadata': _source_metadata
                    }

    def get_file_identifier(self):
        file_list_generator = self.get_stored_file_list()

        file_list_generator = list(file_list_generator)

        if file_list_generator:
            return random.choice(seq=file_list_generator)



from mayan.apps.rest_api import generics
from mayan.apps.smart_settings.permissions import permission_settings_view

from ..models import StoredDriver
from ..serializers.file_metadata_serializers import StoredDriverSerializer


class APIStoredDriverDetailView(generics.RetrieveAPIView):
    """
    get: Returns the details of the selected file metadata driver.
    """
    lookup_url_kwarg = 'stored_driver_id'
    mayan_object_permission_map = {'GET': permission_settings_view}
    model = StoredDriver
    serializer_class = StoredDriverSerializer


class APIStoredDriverListView(generics.ListAPIView):
    """
    get: Returns a list of file metadata drivers.
    """
    mayan_object_permission_map = {'GET': permission_settings_view}
    model = StoredDriver
    serializer_class = StoredDriverSerializer



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('logging', '0002_auto_20211201_1055')
    ]

    operations = [
        migrations.AlterField(
            model_name='storederrorlog',
            name='name',
            field=models.CharField(
                max_length=128, unique=True, verbose_name='Internal name'
            )
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_c

queue_statistics = CeleryQueue(
    label=_(message='Statistics'), name='statistics', transient=True,
    worker=worker_c
)

task_execute_statistic = queue_statistics.add_task_type(
    label=_(message='Execute statistic'),
    dotted_path='mayan.apps.mayan_statistics.tasks.task_execute_statistic'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_DOWNLOAD_FILE_EXPIRATION_INTERVAL,
    DEFAULT_SHARED_UPLOADED_FILE_EXPIRATION_INTERVAL,
    DEFAULT_STORAGE_DOWNLOAD_FILE_STORAGE,
    DEFAULT_STORAGE_DOWNLOAD_FILE_STORAGE_ARGUMENTS,
    DEFAULT_STORAGE_SHARED_STORAGE, DEFAULT_STORAGE_SHARED_STORAGE_ARGUMENTS,
    DEFAULT_STORAGE_TEMPORARY_DIRECTORY
)

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Storage'), name='storage'
)

setting_download_file_expiration_interval = setting_namespace.do_setting_add(
    default=DEFAULT_DOWNLOAD_FILE_EXPIRATION_INTERVAL,
    global_name='DOWNLOAD_FILE_EXPIRATION_INTERVAL', help_text=_(
        message='Time in seconds, after which download files will be deleted.'
    )
)
setting_download_file_storage = setting_namespace.do_setting_add(
    default=DEFAULT_STORAGE_DOWNLOAD_FILE_STORAGE,
    global_name='STORAGE_DOWNLOAD_FILE_STORAGE', help_text=_(
        message='A storage backend that all workers can use to generate and hold '
        'files for download.'
    )
)
setting_download_file_storage_arguments = setting_namespace.do_setting_add(
    default=DEFAULT_STORAGE_DOWNLOAD_FILE_STORAGE_ARGUMENTS,
    global_name='STORAGE_DOWNLOAD_FILE_STORAGE_ARGUMENTS',
)
setting_shared_storage = setting_namespace.do_setting_add(
    default=DEFAULT_STORAGE_SHARED_STORAGE,
    global_name='STORAGE_SHARED_STORAGE', help_text=_(
        message='A storage backend that all workers can use to share files.'
    )
)
setting_shared_storage_arguments = setting_namespace.do_setting_add(
    default=DEFAULT_STORAGE_SHARED_STORAGE_ARGUMENTS,
    global_name='STORAGE_SHARED_STORAGE_ARGUMENTS'
)
setting_temporary_directory = setting_namespace.do_setting_add(
    default=DEFAULT_STORAGE_TEMPORARY_DIRECTORY,
    global_name='STORAGE_TEMPORARY_DIRECTORY', help_text=_(
        message='Temporary directory used site wide to store thumbnails, previews '
        'and temporary files.'
    )
)
setting_shared_uploaded_file_expiration_interval = setting_namespace.do_setting_add(
    default=DEFAULT_SHARED_UPLOADED_FILE_EXPIRATION_INTERVAL,
    global_name='SHARED_UPLOADED_FILE_EXPIRATION_INTERVAL', help_text=_(
        message='Time in seconds, after which temporary uploaded files will be '
        'deleted.'
    )
)



from django.core import management

from ...models import Cache


class Command(management.BaseCommand):
    help = 'Purge the selected cache directly.'
    missing_args_message = 'You must provide a defined storage name that corresponds to a cache.'

    def add_arguments(self, parser):
        parser.add_argument(
            dest='storage_name', help='Name of the cache defined storage.',
            metavar='<storage name>'
        )

    def handle(self, storage_name, **options):
        try:
            cache = Cache.objects.get(defined_storage_name=storage_name)
        except Cache.DoesNotExist:
            self.stderr.write(
                msg='Unknown cache storage `{}`'.format(storage_name)
            )
            exit(1)

        self.stdout.write(
            msg='\nPurging cache "{}"...'.format(cache)
        )

        cache.purge(user=None)

        self.stdout.write(
            msg='\nCache "{}", purged.'.format(cache)
        )



import logging

import extract_msg

from django.utils.translation import gettext_lazy as _

from mayan.apps.file_metadata.classes import FileMetadataDriver
from mayan.apps.storage.literals import MSG_MIME_TYPES

logger = logging.getLogger(__name__)


class FileMetadataDriverExtractMSGTool(FileMetadataDriver):
    description = _(
        message='Extracts information from emails saved in Microsoft '
        'Outlook\'s .msg files.'
    )
    dotted_path_previous_list = (
        'mayan.apps.file_metadata.drivers.extract_msg.ExtractMSGToolDriver',
    )
    internal_name = 'extract_msg'
    label = _(message='Extract msg')
    mime_type_list = list(MSG_MIME_TYPES)

    def _process(self, document_file):
        message = extract_msg.Message(
            path=document_file.open()
        )
        result = {}

        if message.cc:
            result['cc'] = message.cc

        if message.sender:
            result['sender'] = message.sender

        if message.subject:
            result['subject'] = message.subject

        if message.to:
            result['to'] = message.to

        logger.debug('message file metadata: %s', result)

        return result



import platform

if platform.system() in ('FreeBSD', 'OpenBSD', 'Darwin'):
    DEFAULT_DOCUMENT_PARSING_PDFTOTEXT_PATH = '/usr/local/bin/pdftotext'
else:
    DEFAULT_DOCUMENT_PARSING_PDFTOTEXT_PATH = '/usr/bin/pdftotext'

DEFAULT_DOCUMENT_PARSING_AUTO_PARSING = True

ERROR_LOG_DOMAIN_NAME = 'document_parsing'



import logging

from django.core.files import File
from django.db import models
from django.utils.translation import gettext_lazy as _

from mayan.apps.django_gpg.exceptions import DecryptionError
from mayan.apps.django_gpg.models import Key
from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.storage.utils import NamedTemporaryFile

from .events import (
    event_detached_signature_created, event_embedded_signature_created
)

logger = logging.getLogger(name=__name__)


class DetachedSignatureManager(models.Manager):
    def sign_document_file(
        self, document_file, key, passphrase=None, user=None
    ):
        with NamedTemporaryFile() as temporary_file_object:
            with document_file.open() as file_object:
                key.sign_file(
                    binary=True, detached=True, file_object=file_object,
                    output=temporary_file_object.name,
                    passphrase=passphrase
                )
            temporary_file_object.seek(0)

            instance = self.model(
                document_file=document_file,
                signature_file=File(file=temporary_file_object)
            )
            instance._event_ignore = True
            instance.save()

            # Release the descriptor opened by Django when assigning the
            # `signature_file` field content.
            instance.signature_file.file.close()

            event_detached_signature_created.commit(
                action_object=instance, actor=user, target=document_file
            )

        return instance


class EmbeddedSignatureManager(models.Manager):
    def open_signed(self, document_file, file_object):
        for signature in self.filter(document_file=document_file):
            try:
                return self.open_signed(
                    file_object=Key.objects.decrypt_file(
                        file_object=file_object
                    ), document_file=document_file
                )
            except DecryptionError:
                file_object.seek(0)
                return file_object
        else:
            return file_object

    def sign_document_file(
        self, document_file, key, passphrase=None, user=None
    ):
        with NamedTemporaryFile() as temporary_file_object:
            try:
                with document_file.open() as file_object:
                    key.sign_file(
                        binary=True, file_object=file_object,
                        output=temporary_file_object.name,
                        passphrase=passphrase
                    )
            except Exception:
                raise
            else:
                # The result of key.sign_file does not contain the
                # signature ID.
                # Verify the signed file to obtain the signature ID.
                temporary_file_object.seek(0)
                result = Key.objects.verify_file(
                    file_object=temporary_file_object
                )

                # Reset the file pointer and use it to create the new
                # signed document file.
                temporary_file_object.seek(0)
                document_file.document.files_upload(
                    file_object=temporary_file_object,
                    filename='{}_{}'.format(
                        str(document_file), _(message='signed')
                    ), user=user
                )
                instance = self.get(signature_id=result.signature_id)
                event_embedded_signature_created.commit(
                    action_object=instance, actor=user, target=document_file
                )
                return instance

    def unsigned_document_files(self):
        return DocumentFile.objects.exclude(
            pk__in=self.values('document_file')
        )



import re

from django.template import Library

from ..utils import process_regex_flags

register = Library()


@register.simple_tag(name='regex_findall')
def tag_regex_findall(pattern, string, **kwargs):
    """
    Return all non-overlapping matches of pattern in string, as a list of
    strings. {% regex_findall pattern string flags %}
    """
    flags = process_regex_flags(**kwargs)
    return re.findall(flags=flags, pattern=pattern, string=string)


@register.simple_tag(name='regex_match')
def tag_regex_match(pattern, string, **kwargs):
    """
    If zero or more characters at the beginning of string match the regular
    expression pattern, return a corresponding match object.
    {% regex_match pattern string flags %}
    """
    flags = process_regex_flags(**kwargs)
    return re.match(flags=flags, pattern=pattern, string=string)


@register.simple_tag(name='regex_search')
def tag_regex_search(pattern, string, **kwargs):
    """
    Scan through string looking for the first location where the regular
    expression pattern produces a match, and return a corresponding
    match object. {% regex_search pattern string flags %}
    """
    flags = process_regex_flags(**kwargs)
    return re.search(flags=flags, pattern=pattern, string=string)


@register.simple_tag(name='regex_sub')
def tag_regex_sub(pattern, repl, string, count=0, **kwargs):
    """
    Replacing the leftmost non-overlapping occurrences of pattern in
    string with repl. {% regex_sub pattern repl string count=0 flags %}
    """
    flags = process_regex_flags(**kwargs)
    return re.sub(
        count=count, flags=flags, pattern=pattern, repl=repl, string=string
    )



import pyotp

from django.contrib import messages
from django.core.signing import BadSignature, dumps, loads
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import RedirectView

from mayan.apps.views.generics import (
    ConfirmView, FormView, SingleObjectDetailView
)
from mayan.apps.views.http import URL

from .forms import FormUserOTPDataDetail, FormUserOTPDataEdit
from .view_mixins import OTPBackendEnabledViewMixin


class UserOTPDataDetailView(
    OTPBackendEnabledViewMixin, SingleObjectDetailView
):
    form_class = FormUserOTPDataDetail

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _(
                message='One time pad details for user: %s'
            ) % self.object
        }

    def get_object(self):
        return self.request.user


class UserOTPDataDisableView(OTPBackendEnabledViewMixin, ConfirmView):
    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().dispatch(request=request, *args, **kwargs)

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _(
                message='Disable one time pad for user: %s'
            ) % self.object
        }

    def get_object(self):
        return self.request.user

    def view_action(self):
        self.request.user.otp_data.disable()

        messages.success(
            message=_(message='OTP disable successfully.'), request=self.request
        )


class UserOTPDataEnableView(OTPBackendEnabledViewMixin, RedirectView):
    def get_redirect_url(self, *args, **kwargs):
        signed_secret = dumps(
            obj=pyotp.random_base32()
        )

        return URL(
            path=reverse(viewname='authentication_otp:otp_verify'),
            query={'signed_secret': signed_secret}
        ).to_string()


class UserOTPDataVerifyTokenView(OTPBackendEnabledViewMixin, FormView):
    form_class = FormUserOTPDataEdit

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()

        if request.user.otp_data.is_enabled():
            messages.info(
                message=_(message='OTP is already enabled.'),
                request=self.request
            )
            return HttpResponseRedirect(
                redirect_to=reverse(
                    viewname='authentication_otp:otp_detail'
                )
            )
        else:
            return super().dispatch(request=request, *args, **kwargs)

    def form_valid(self, form):
        secret = form.cleaned_data['secret']
        token = form.cleaned_data['token']
        self.request.user.otp_data.enable(secret=secret, token=token)

        messages.success(
            message=_(message='OTP enabled successfully.'),
            request=self.request
        )

        return super().form_valid(form=form)

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _(
                message='Enable one time pad for user: %s'
            ) % self.object
        }

    def get_form_extra_kwargs(self):
        signed_secret = self.request.GET.get('signed_secret', '')

        try:
            secret = loads(s=signed_secret)
        except BadSignature:
            messages.error(
                message=_(
                    message='OTP secret validation error.'
                ), request=self.request
            )
            secret = None

        return {
            'initial': {
                'secret': secret,
                'signed_secret': signed_secret
            },
            'user': self.request.user
        }

    def get_object(self):
        return self.request.user



import logging

import celery

from django.apps import apps
from django.contrib.auth import get_user_model

from mayan.apps.lock_manager.exceptions import LockError
from mayan.celery import app

from .settings import setting_image_generation_max_retries
from .utils import IndexedDictionary

logger = logging.getLogger(name=__name__)


@app.task(
    bind=True, max_retries=setting_image_generation_max_retries.value,
    retry_backoff=True
)
def task_content_object_image_generate(
    self, content_type_id, object_id, maximum_layer_order=None,
    transformation_dictionary_list=None, user_id=None
):
    ContentType = apps.get_model(
        app_label='contenttypes', model_name='ContentType'
    )
    User = get_user_model()

    content_type = ContentType.objects.get(pk=content_type_id)

    if user_id:
        user = User.objects.get(pk=user_id)
    else:
        user = None

    obj = content_type.get_object_for_this_type(pk=object_id)

    transformation_indexed_dictionary = IndexedDictionary.from_dictionary_list(
        dictionary_list=transformation_dictionary_list or ()
    )

    transformation_instance_list = transformation_indexed_dictionary.as_instance_list()

    try:
        return obj.generate_image(
            maximum_layer_order=maximum_layer_order,
            transformation_instance_list=transformation_instance_list,
            user=user
        )
    except LockError as exception:
        logger.warning(
            'LockError during attempt to generate image for %s. Retrying.',
            obj
        )
        try:
            raise self.retry(exc=exception)
        except celery.exceptions.MaxRetriesExceededError:
            logger.error(
                'Maximum retries reached for image generation task. '
                'System might be overloaded or a stale lock might be '
                'preventing the task from completing.'
            )
            raise



from django.db import migrations

PATH_MAPPING = {
    'mayan.apps.credentials_google.credential_backends.CredentialBackendGoogleServiceAccount': 'mayan.apps.credentials.credential_backends.CredentialBackendGoogleServiceAccount'
}


def code_path_update(apps, schema_editor):
    StoredCredential = apps.get_model(
        app_label='credentials', model_name='StoredCredential'
    )

    for key, value in PATH_MAPPING.items():
        queryset = StoredCredential.objects.using(
            alias=schema_editor.connection.alias
        ).filter(backend_path=key)

        queryset.update(backend_path=value)


def reverse_code_path_update(apps, schema_editor):
    StoredCredential = apps.get_model(
        app_label='credentials', model_name='StoredCredential'
    )

    for key, value in PATH_MAPPING.items():
        queryset = StoredCredential.objects.using(
            alias=schema_editor.connection.alias
        ).filter(backend_path=value)

        queryset.update(backend_path=key)


class Migration(migrations.Migration):
    dependencies = [
        ('credentials', '0005_auto_20210207_0840')
    ]

    operations = [
        migrations.RunPython(
            code=code_path_update, reverse_code=reverse_code_path_update
        )
    ]
